# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: job.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import dag_pb2 as dag__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\tjob.proto\x12\x04jobs\x1a\tdag.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\x98\x07\n\x03Job\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12.\n\ncreated_at\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x1c\n\x04\x63ode\x18\x05 \x01(\x0b\x32\x0e.jobs.Job.Code\x12 \n\x08resource\x18\x06 \x01(\x0b\x32\x0e.jobs.Resource\x12\x32\n\rfeature_gates\x18\x07 \x03(\x0b\x32\x1b.jobs.Job.FeatureGatesEntry\x12$\n\tauth_info\x18\x0b \x01(\x0b\x32\x11.jobs.NBXAuthInfo\x12$\n\x08schedule\x18\x0c \x01(\x0b\x32\x12.jobs.Job.Schedule\x12\x1a\n\x03\x64\x61g\x18\r \x01(\x0b\x32\r.jobs.dag.DAG\x12\x0e\n\x06paused\x18\x0e \x01(\x08\x12 \n\x06status\x18\x0f \x01(\x0e\x32\x10.jobs.Job.Status\x12\x1e\n\x07k8sInfo\x18\x10 \x01(\x0b\x32\r.jobs.K8sInfo\x1a\xd7\x01\n\x04\x43ode\x12\x0e\n\x06s3_url\x18\x01 \x01(\t\x12\x0c\n\x04size\x18\x02 \x01(\x05\x12!\n\x04type\x18\x03 \x01(\x0e\x32\x13.jobs.Job.Code.Type\x12+\n\x07s3_meta\x18\x04 \x03(\x0b\x32\x1a.jobs.Job.Code.S3MetaEntry\x1a-\n\x0bS3MetaEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"2\n\x04Type\x12\x0b\n\x07NOT_SET\x10\x00\x12\x08\n\x04NBOX\x10\x01\x12\x07\n\x03ZIP\x10\x02\x12\n\n\x06\x42INARY\x10\x03\x1a\x33\n\x11\x46\x65\x61tureGatesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1al\n\x08Schedule\x12)\n\x05start\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\'\n\x03\x65nd\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0c\n\x04\x63ron\x18\x03 \x01(\t\"\x9b\x01\n\x06Status\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07NOT_SET\x10\x01\x12\r\n\tPREPARING\x10\x02\x12\n\n\x06PAUSED\x10\x03\x12\r\n\tSCHEDULED\x10\x04\x12\n\n\x06\x41\x43TIVE\x10\x05\x12\t\n\x05\x45RROR\x10\x06\x12\r\n\tCOMPLETED\x10\x07\x12\x0c\n\x08\x41RCHIVED\x10\x08\x12\r\n\tCANCELLED\x10\t\x12\n\n\x06KILLED\x10\n\"v\n\x0bNBXAuthInfo\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x14\n\x0cworkspace_id\x18\x02 \x01(\t\x12\x14\n\x0c\x61\x63\x63\x65ss_token\x18\x03 \x01(\t\x12\x15\n\rrefresh_token\x18\x04 \x01(\t\x12\x12\n\naccess_key\x18\x05 \x01(\t\")\n\x07K8sInfo\x12\x10\n\x08\x63ron_job\x18\x01 \x01(\t\x12\x0c\n\x04pods\x18\x02 \x03(\t\"\x80\x01\n\x08Resource\x12\x0b\n\x03\x63pu\x18\x01 \x01(\t\x12\x0e\n\x06memory\x18\x02 \x01(\t\x12\x0b\n\x03gpu\x18\x03 \x01(\t\x12\x11\n\tgpu_count\x18\x04 \x01(\t\x12\x11\n\tdisk_size\x18\x05 \x01(\t\x12\x0f\n\x07timeout\x18\x06 \x01(\x03\x12\x13\n\x0bmax_retries\x18\x07 \x01(\x05\x42\x32Z0github.com/NimbleBoxAI/jobs-architecture/jobs_pbb\x06proto3')



_JOB = DESCRIPTOR.message_types_by_name['Job']
_JOB_CODE = _JOB.nested_types_by_name['Code']
_JOB_CODE_S3METAENTRY = _JOB_CODE.nested_types_by_name['S3MetaEntry']
_JOB_FEATUREGATESENTRY = _JOB.nested_types_by_name['FeatureGatesEntry']
_JOB_SCHEDULE = _JOB.nested_types_by_name['Schedule']
_NBXAUTHINFO = DESCRIPTOR.message_types_by_name['NBXAuthInfo']
_K8SINFO = DESCRIPTOR.message_types_by_name['K8sInfo']
_RESOURCE = DESCRIPTOR.message_types_by_name['Resource']
_JOB_CODE_TYPE = _JOB_CODE.enum_types_by_name['Type']
_JOB_STATUS = _JOB.enum_types_by_name['Status']
Job = _reflection.GeneratedProtocolMessageType('Job', (_message.Message,), {

  'Code' : _reflection.GeneratedProtocolMessageType('Code', (_message.Message,), {

    'S3MetaEntry' : _reflection.GeneratedProtocolMessageType('S3MetaEntry', (_message.Message,), {
      'DESCRIPTOR' : _JOB_CODE_S3METAENTRY,
      '__module__' : 'job_pb2'
      # @@protoc_insertion_point(class_scope:jobs.Job.Code.S3MetaEntry)
      })
    ,
    'DESCRIPTOR' : _JOB_CODE,
    '__module__' : 'job_pb2'
    # @@protoc_insertion_point(class_scope:jobs.Job.Code)
    })
  ,

  'FeatureGatesEntry' : _reflection.GeneratedProtocolMessageType('FeatureGatesEntry', (_message.Message,), {
    'DESCRIPTOR' : _JOB_FEATUREGATESENTRY,
    '__module__' : 'job_pb2'
    # @@protoc_insertion_point(class_scope:jobs.Job.FeatureGatesEntry)
    })
  ,

  'Schedule' : _reflection.GeneratedProtocolMessageType('Schedule', (_message.Message,), {
    'DESCRIPTOR' : _JOB_SCHEDULE,
    '__module__' : 'job_pb2'
    # @@protoc_insertion_point(class_scope:jobs.Job.Schedule)
    })
  ,
  'DESCRIPTOR' : _JOB,
  '__module__' : 'job_pb2'
  # @@protoc_insertion_point(class_scope:jobs.Job)
  })
_sym_db.RegisterMessage(Job)
_sym_db.RegisterMessage(Job.Code)
_sym_db.RegisterMessage(Job.Code.S3MetaEntry)
_sym_db.RegisterMessage(Job.FeatureGatesEntry)
_sym_db.RegisterMessage(Job.Schedule)

NBXAuthInfo = _reflection.GeneratedProtocolMessageType('NBXAuthInfo', (_message.Message,), {
  'DESCRIPTOR' : _NBXAUTHINFO,
  '__module__' : 'job_pb2'
  # @@protoc_insertion_point(class_scope:jobs.NBXAuthInfo)
  })
_sym_db.RegisterMessage(NBXAuthInfo)

K8sInfo = _reflection.GeneratedProtocolMessageType('K8sInfo', (_message.Message,), {
  'DESCRIPTOR' : _K8SINFO,
  '__module__' : 'job_pb2'
  # @@protoc_insertion_point(class_scope:jobs.K8sInfo)
  })
_sym_db.RegisterMessage(K8sInfo)

Resource = _reflection.GeneratedProtocolMessageType('Resource', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCE,
  '__module__' : 'job_pb2'
  # @@protoc_insertion_point(class_scope:jobs.Resource)
  })
_sym_db.RegisterMessage(Resource)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z0github.com/NimbleBoxAI/jobs-architecture/jobs_pb'
  _JOB_CODE_S3METAENTRY._options = None
  _JOB_CODE_S3METAENTRY._serialized_options = b'8\001'
  _JOB_FEATUREGATESENTRY._options = None
  _JOB_FEATUREGATESENTRY._serialized_options = b'8\001'
  _JOB._serialized_start=64
  _JOB._serialized_end=984
  _JOB_CODE._serialized_start=448
  _JOB_CODE._serialized_end=663
  _JOB_CODE_S3METAENTRY._serialized_start=566
  _JOB_CODE_S3METAENTRY._serialized_end=611
  _JOB_CODE_TYPE._serialized_start=613
  _JOB_CODE_TYPE._serialized_end=663
  _JOB_FEATUREGATESENTRY._serialized_start=665
  _JOB_FEATUREGATESENTRY._serialized_end=716
  _JOB_SCHEDULE._serialized_start=718
  _JOB_SCHEDULE._serialized_end=826
  _JOB_STATUS._serialized_start=829
  _JOB_STATUS._serialized_end=984
  _NBXAUTHINFO._serialized_start=986
  _NBXAUTHINFO._serialized_end=1104
  _K8SINFO._serialized_start=1106
  _K8SINFO._serialized_end=1147
  _RESOURCE._serialized_start=1150
  _RESOURCE._serialized_end=1278
# @@protoc_insertion_point(module_scope)



"""
All the operators related to communication are here.
"""

from nbox import Operator

class Notify(Operator):
  _mode_to_packages = {
    "slack": "slackclient",
    "ms_teams": "microsoft-teams",
    "discord": "discord",
  }

  def __init__(
    self,
    slack_connect: str = None,
    ms_teams: str = None,
    discord: str = None,
  ):
    """Notifications"""
    super().__init__()
    self.notify_mode = None
    self.notify_id = None

    for mode, id in [
      ("slack", slack_connect),
      ("ms_teams", ms_teams),
      ("discord", discord),
    ]:
      if id:
        self.notify_mode = mode
        self.notify_id = id

        # check for package dependencies
        import importlib
        try:
          importlib.import_module(self._mode_to_packages[mode])
        except ImportError:
          raise Exception(f"{self._mode_to_packages[mode]} package required for {mode}")
        break
    
    if not self.notify_mode:
      raise Exception("No notification mode specified")

  def forward(self, message: str, **kwargs):
    package = self._mode_to_packages[self.notify_mode]
    import importlib
    importlib.import_module(package)
    if self.notify_mode == "slack":
      from slackclient import SlackClient
      sc = SlackClient(self.notify_id)
      sc.api_call("chat.postMessage", text = message, **kwargs)
    elif self.notify_mode == "ms_teams":
      from microsoft_teams.api_client import TeamsApiClient
      from microsoft_teams.models import MessageCard
      client = TeamsApiClient(self.notify_id)
      client.connect()
      client.send_message(MessageCard(text = message, **kwargs))
    elif self.notify_mode == "discord":
      from discord import Webhook, RequestsWebhookAdapter
      webhook = Webhook.from_url(self.notify_id, adapter = RequestsWebhookAdapter())
      webhook.send(message, **kwargs)




import io
import sys
import base64
import logging
import traceback
from json import loads
from inspect import signature

import requests
import random
from functools import partial
from google.protobuf.json_format import ParseDict, MessageToDict, MessageToJson
from google.protobuf.text_format import MessageToString, Parse as ParseString
from google.protobuf.timestamp_pb2 import Timestamp
from google.protobuf.message import Message

from .rest_pb2 import Echo

from nbox.utils import logger as _logger

message_to_dict = partial(
  MessageToDict,
  including_default_value_fields = True,
  preserving_proto_field_name = True,
  use_integers_for_enums = True,
  float_precision = 4
)

dict_to_message = partial(
  ParseDict,
  ignore_unknown_fields = True,
)

message_to_json = partial(
  MessageToJson,
  including_default_value_fields = True,
  preserving_proto_field_name = True,
  use_integers_for_enums = True,
  float_precision = 4
)

message_to_string = partial(
  MessageToString,
  as_utf8 = False,
  as_one_line = False, # does a simple rstrip on the end
  # use_short_repeated_primitives = False,
  pointy_brackets = False,
  use_index_order = False,
  use_field_number=False,
  print_unknown_fields=True,
)

string_to_message = partial(
  ParseString,
  allow_unknown_field = True,
)

def message_to_b64(message: Message) -> str:
  return base64.b64encode(message.SerializeToString()).decode("utf-8")

def b64_to_message(b64: str, message: Message) -> Message:
  message.ParseFromString(base64.b64decode(b64))
  return message


def call_rpc(sess: requests.Session, url: str, message: Echo = None):
  fn = sess.post
  if message != None:
    message = message_to_dict(message)
  # _logger.info(f"POST {url} | json: {message}")
  r = fn(url, json = message)

  if r.status_code == 400:
    data = loads(r.text)
    out = dict_to_message(data, Echo())
    if out.base64_string == "":
      _logger.error(f"400: {out.message.strip()}")
      return None
  elif r.status_code == 501:
    _logger.error("501: NOT IMPLEMENTED")
    return None
  elif r.status_code == 500:
    _logger.error("500: INTERNAL SERVER ERROR")
    return None
  elif r.status_code != 200:
    _logger.error(f"Unknown error: {r.status_code}")
    _logger.error(r.text)
    return None
    
  out = dict_to_message(r.json(), Echo())
  return out

def run_rpc(service_fn, message, **kwargs) -> Message:
  _echo = default_echo()

  # check if correct arguments are being sent or not
  args = signature(service_fn).parameters
  if len(args) - 1 < len(kwargs):
    # this is a 500 because client can never send more than this can take in
    _logger.error(f"Service function takes upto {len(args) - 1} arguments, but {len(kwargs)} were provided")
    _echo.message = "500: INTERNAL SERVER ERROR"
    return _echo

  # run it and handle any exceptions
  try:
    response = service_fn(message, **kwargs)
  except NotImplementedError:
    return Echo(server_time=get_timestamp(), message = "501: NOT IMPLEMENTED")
  except Exception as e:
    f = traceback.format_exc()
    for l in f.splitlines():
      _logger.error(l)
    return Echo(server_time=get_timestamp(), message = "500: INTERNAL SERVER ERROR")
  
  # manage response
  if isinstance(response, str):
    _echo.message = f"400: {response}" 
    response = _echo
  return response

def get_timestamp() -> Timestamp:
  _time = Timestamp()
  _time.GetCurrentTime()
  return _time

def default_echo() -> Echo:
  return Echo(
    server_time=get_timestamp(),
    message = "".join(random.choices("abcedfghijklmnopqrtsuvwxyz0123456789", k = 32))
  )



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import proto.relics_pb2
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class CreateRelicRequest(google.protobuf.message.Message):
    """things related to general WebAPIs"""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    REGION_FIELD_NUMBER: builtins.int
    workspace_id: typing.Text
    """the main separation layer"""

    name: typing.Text
    region: typing.Text
    def __init__(self,
        *,
        workspace_id: typing.Text = ...,
        name: typing.Text = ...,
        region: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["name",b"name","region",b"region","workspace_id",b"workspace_id"]) -> None: ...
global___CreateRelicRequest = CreateRelicRequest

class ListRelicsRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    RELIC_NAME_FIELD_NUMBER: builtins.int
    PAGE_NO_FIELD_NUMBER: builtins.int
    workspace_id: typing.Text
    """the main separation layer"""

    relic_name: typing.Text
    """the name of the relic"""

    page_no: builtins.int
    """the page number for pagination needs"""

    def __init__(self,
        *,
        workspace_id: typing.Text = ...,
        relic_name: typing.Text = ...,
        page_no: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["page_no",b"page_no","relic_name",b"relic_name","workspace_id",b"workspace_id"]) -> None: ...
global___ListRelicsRequest = ListRelicsRequest

class ListRelicsResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    RELICS_FIELD_NUMBER: builtins.int
    TOTAL_RELICS_FIELD_NUMBER: builtins.int
    @property
    def relics(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[proto.relics_pb2.Relic]: ...
    total_relics: builtins.int
    """the total number of relics used for pagination"""

    def __init__(self,
        *,
        relics: typing.Optional[typing.Iterable[proto.relics_pb2.Relic]] = ...,
        total_relics: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["relics",b"relics","total_relics",b"total_relics"]) -> None: ...
global___ListRelicsResponse = ListRelicsResponse

class ListRelicFilesRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    RELIC_ID_FIELD_NUMBER: builtins.int
    RELIC_NAME_FIELD_NUMBER: builtins.int
    PREFIX_FIELD_NUMBER: builtins.int
    FILE_NAME_FIELD_NUMBER: builtins.int
    PAGE_NO_FIELD_NUMBER: builtins.int
    workspace_id: typing.Text
    """the main separation layer"""

    relic_id: typing.Text
    """the id of the relic"""

    relic_name: typing.Text
    """the id of the relic"""

    prefix: typing.Text
    """the prefix for the files"""

    file_name: typing.Text
    """for searching through the files, we may not support this functionality
    right now, but this is a crucial system to have
    """

    page_no: builtins.int
    """the page number for pagination needs"""

    def __init__(self,
        *,
        workspace_id: typing.Text = ...,
        relic_id: typing.Text = ...,
        relic_name: typing.Text = ...,
        prefix: typing.Text = ...,
        file_name: typing.Text = ...,
        page_no: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["file_name",b"file_name","page_no",b"page_no","prefix",b"prefix","relic_id",b"relic_id","relic_name",b"relic_name","workspace_id",b"workspace_id"]) -> None: ...
global___ListRelicFilesRequest = ListRelicFilesRequest

class ListRelicFilesResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    FILES_FIELD_NUMBER: builtins.int
    TOTAL_FILES_FIELD_NUMBER: builtins.int
    @property
    def files(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[proto.relics_pb2.RelicFile]: ...
    total_files: builtins.int
    """the total number of files used for pagination"""

    def __init__(self,
        *,
        files: typing.Optional[typing.Iterable[proto.relics_pb2.RelicFile]] = ...,
        total_files: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["files",b"files","total_files",b"total_files"]) -> None: ...
global___ListRelicFilesResponse = ListRelicFilesResponse



"""
This is the code for NBX-Relics which is a simple file system for your organisation.
"""
import os
import cloudpickle
import requests
import tabulate
from hashlib import md5
from typing import List
from copy import deepcopy
from functools import lru_cache

from nbox.auth import secret
from nbox.init import nbox_ws_v1
from nbox.utils import logger, env
from nbox.sublime.relics_rpc_client import (
  RelicStore_Stub,
  RelicFile,
  Relic as RelicProto,
  CreateRelicRequest,
  ListRelicFilesRequest,
  ListRelicsRequest
)
from nbox.relics.base import BaseStore
from nbox.auth import ConfigString, secret

def get_relic_file(fpath: str, username: str, workspace_id: str):
  # assert os.path.exists(fpath), f"File {fpath} does not exist"
  # assert os.path.isfile(fpath), f"File {fpath} is not a file"

  # clean up fpath, remove any trailing slashes
  # trim any . or / from prefix and suffix
  fpath_cleaned = fpath.strip("./")

  extra = {}
  if os.path.exists(fpath):
    file_stat = os.stat(fpath)
    extra = {
      "created_on": int(file_stat.st_mtime),    # int
      "last_modified": int(file_stat.st_mtime), # int
      "size": max(1, file_stat.st_size),        # bytes
    }
  return RelicFile(
    name = fpath_cleaned,
    username = username,
    type = RelicFile.RelicType.FILE,
    workspace_id = workspace_id,
    **extra
  )


@lru_cache()
def _get_stub():
  # url = "http://0.0.0.0:8081/relics" # debug
  url = secret.get("nbx_url") + "/relics"
  logger.debug("Connecting to RelicStore at: " + url)
  session = deepcopy(nbox_ws_v1._session)
  stub = RelicStore_Stub(url, session)
  return stub


def print_relics(workspace_id: str = ""):
  stub = _get_stub()
  workspace_id = workspace_id or secret.get(ConfigString.workspace_id.value)
  req = ListRelicsRequest(workspace_id = workspace_id,)
  out = stub.list_relics(req)
  headers = ["relic_name",]
  rows = [[r.name,] for r in out.relics]
  for l in tabulate.tabulate(rows, headers).splitlines():
    logger.info(l)


class RelicsNBX(BaseStore):
  list = staticmethod(print_relics)

  def __init__(self, relic_name: str, workspace_id: str = "", create: bool = False, prefix: str = ""):
    """
    The client for NBX-Relics.

    Args:
      relic_name (str): The name of the relic.
      workspace_id (str): The workspace ID, if not provided, will be one in global config.
      create (bool): Create the relic if it does not exist.
      prefix (str): The prefix to use for all files in this relic. If provided all the files are uploaded and downloaded with this prefix.
    """
    self.workspace_id = workspace_id or secret.get(ConfigString.workspace_id.value)
    self.relic_name = relic_name
    self.username = secret.get("username") # if its in the job then this part will automatically be filled
    self.prefix = prefix.strip("/")
    self.stub = _get_stub()
    _relic = self.stub.get_relic_details(RelicProto(workspace_id=self.workspace_id, name=relic_name,))
    # print("asdfasdfasdfasdf", _relic, not _relic and create)
    if not _relic and create:
      # this means that a new one will have to be created
      logger.debug(f"Creating new relic {relic_name}")
      self.relic = self.stub.create_relic(CreateRelicRequest(workspace_id=self.workspace_id, name = relic_name,))
      logger.debug(f"Created new relic {self.relic}")
    else:
      self.relic = _relic

  def __repr__(self):
    return f"RelicStore({self.workspace_id}, {self.relic_name}, {'CONNECTED' if self.relic else 'NOT CONNECTED'})"

  def _upload_relic_file(self, local_path: str, relic_file: RelicFile):
    if not relic_file.relic_name:
      raise ValueError("relic_name not set in RelicFile")
    if self.prefix:
      relic_file.name = f"{self.prefix}/{relic_file.name}"

    # ideally this is a lot like what happens in nbox
    logger.debug(f"Uploading {local_path} to {relic_file.name}")
    out = self.stub.create_file(_RelicFile = relic_file,)
    if not out.url:
      raise Exception("Could not get link")
    
    # do not perform merge here because "url" might get stored in MongoDB
    # relic_file.MergeFrom(out)
    r = requests.post(
      url = out.url,
      data = out.body,
      files={
        "file": (out.body["key"], open(local_path, "rb"))
      }
    )
    logger.debug(f"Upload status: {r.status_code}")
    r.raise_for_status()

  def _download_relic_file(self, local_path: str, relic_file: RelicFile):
    if self.relic is None:
      raise ValueError("Relic does not exist, pass create=True")
    if self.prefix:
      relic_file.name = self.prefix + "/" + relic_file.name

    # ideally this is a lot like what happens in nbox
    logger.debug(f"Downloading {local_path} from S3 ...")
    out = self.stub.download_file(_RelicFile = relic_file,)
    if not out.url:
      raise Exception("Could not get link, are you sure this file exists?")
    
    # do not perform merge here because "url" might get stored in MongoDB
    # relic_file.MergeFrom(out)
    # logger.debug(f"URL: {out.url}")
    with requests.get(out.url, stream=True) as r:
      r.raise_for_status()
      total_size = 0
      with open(local_path, 'wb') as f:
        for chunk in r.iter_content(chunk_size=8192): 
          # If you have chunk encoded response uncomment if
          # and set chunk_size parameter to None.
          #if chunk: 
          f.write(chunk)
          total_size += len(chunk)
    logger.debug(f"Download '{local_path}' status: OK ({total_size//1024} KB)")

  """
  At it's core the Relic is supposed to be a file system and not a client. Thus you cannot download something
  from a relic, but rather you tell the path you want to read and it will return the file. This is because of the
  fact that this is nothing but a glorified key value store.

  Moreover Relic as a broader concept is a storage in Von Neumann architecture. It is a storage that is not, so
  the more ways to store files built into it, the better the experience. So there are different types of put and
  get methods.
  """

  def put(self, local_path: str):
    """Put the file at this path into the relic"""
    if self.relic is None:
      raise ValueError("Relic does not exist, pass create=True")
    logger.debug(f"Putting file: {local_path}")
    relic_file = get_relic_file(local_path, self.username, self.workspace_id)
    relic_file.relic_name = self.relic_name
    self._upload_relic_file(local_path, relic_file)

  def put_to(self, local_path: str, remote_path: str) -> None:
    if self.relic is None:
      raise ValueError("Relic does not exist, pass create=True")
    logger.debug(f"Putting file: {local_path} to {remote_path}")
    relic_file = get_relic_file(local_path, self.username, self.workspace_id)
    relic_file.relic_name = self.relic_name
    relic_file.name = remote_path # override the name
    self._upload_relic_file(local_path, relic_file)

  def get(self, local_path: str):
    """Get the file at this path from the relic"""
    if self.relic is None:
      raise ValueError("Relic does not exist, pass create=True")
    logger.debug(f"Getting file: {local_path}")
    relic_file = RelicFile(name = local_path.strip("./"),)
    relic_file.relic_name = self.relic_name
    relic_file.workspace_id = self.workspace_id
    self._download_relic_file(local_path, relic_file)

  def get_from(self, local_path: str, remote_path: str) -> None:
    if self.relic is None:
      raise ValueError("Relic does not exist, pass create=True")
    logger.debug(f"Getting file: {local_path} from {remote_path}")
    relic_file = RelicFile(name = remote_path.strip("./"),)
    relic_file.relic_name = self.relic_name
    relic_file.workspace_id = self.workspace_id
    self._download_relic_file(local_path, relic_file)

  def rm(self, local_path: str):
    """Delete the file at this path from the relic"""
    if self.relic is None:
      raise ValueError("Relic does not exist, pass create=True")
    logger.debug(f"Getting file: {local_path}")
    relic_file = get_relic_file(local_path, self.username, self.workspace_id)
    relic_file.relic_name = self.relic_name
    out = self.stub.delete_relic_file(relic_file)
    if not out.success:
      logger.error(out.message)
      raise ValueError("Could not delete file")

  def has(self, local_path: str):
    prefix, file_name = os.path.split(local_path)
    out = self.stub.list_relic_files(
      ListRelicFilesRequest(
        workspace_id=self.workspace_id,
        relic_name=self.relic_name,
        prefix=prefix,
        file_name=file_name
      )
    )
    for f in out.files:
      if f.name.strip("/") == local_path.strip("/"):
        return True
    return False

  """
  There are other convinience methods provided to keep consistency between the different types of relics. Note
  that we do no have a baseclass right now because I am note sure what are all the possible features we can have
  in common with all.
  """

  def put_object(self, key: str, py_object):
    """wrapper function for putting a python object"""
    # we will cache the object in the local file system
    cache_dir = os.path.join(env.NBOX_HOME_DIR(), ".cache")
    if not os.path.exists(cache_dir):
      os.makedirs(cache_dir)
    _key = os.path.join(cache_dir, md5(key.encode()).hexdigest())
    with open(_key, "wb") as f:
      cloudpickle.dump(py_object, f)
    self.put_to(_key, key)

  def get_object(self, key: str):
    """wrapper function for getting a python object"""
    cache_dir = os.path.join(env.NBOX_HOME_DIR(), ".cache")
    if not os.path.exists(cache_dir):
      os.makedirs(cache_dir)
    _key = os.path.join(cache_dir, md5(key.encode()).hexdigest())
    self.get_from(_key, key)
    with open(_key, "rb") as f:
      out = cloudpickle.load(f)
    return out

  """
  Some APIs are more on the level of the relic itself.
  """

  def delete(self):
    """Deletes your relic"""
    if self.relic is None:
      raise ValueError("Relic does not exist, nothing to delete")
    logger.warning(f"Deleting relic {self.relic_name}")
    self.stub.delete_relic(self.relic)

  def list_files(self, path: str = "") -> List[RelicFile]:
    """List all the files in the relic at path"""
    if self.relic is None:
      raise ValueError("Relic does not exist, pass create=True")
    logger.debug(f"Listing files in relic {self.relic_name}")
    out = self.stub.list_relic_files(RelicFile(
      workspace_id = self.workspace_id,
      relic_name = self.relic_name,
      name = path
    ))
    return out.files

  def start_fs():
    # /my_relic/.....
    pass

# nbx jobs ... trigger --mount="dataset:/my-dataset/email/,model_master:/model"



"""
gRPC + Protobuf = ♥️

Contains methods defined [here](https://github.com/googleapis/python-api-core/blob/main/google/api_core/protobuf_helpers.py)
"""

# Copyright 2017 Google LLC
# Copyright 2022 NimbleBox.ai
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import inspect
import collections
import collections.abc
from grpc import RpcError

from google.protobuf import message
from google.protobuf import wrappers_pb2
from google.protobuf import field_mask_pb2

from google.protobuf.json_format import MessageToDict, MessageToJson, ParseDict

from nbox.utils import logger

_SENTINEL = object()
_WRAPPER_TYPES = (
  wrappers_pb2.BoolValue,
  wrappers_pb2.BytesValue,
  wrappers_pb2.DoubleValue,
  wrappers_pb2.FloatValue,
  wrappers_pb2.Int32Value,
  wrappers_pb2.Int64Value,
  wrappers_pb2.StringValue,
  wrappers_pb2.UInt32Value,
  wrappers_pb2.UInt64Value,
)

def rpc(stub: callable, message, err_msg: str, raise_on_error: bool = True):
  """convienience function for calling a gRPC stub"""
  try:
    resp = stub(message)
    raise_on_error = False
  except RpcError as e:
    logger.error(err_msg)
    # err_ = loads(e.debug_error_string())
    # if "value" in err_:
    #   if int(err_["value"]) > 500:
    #     logger.error("There is something wrong from our side. Your files are safe on your local machine.")
    #   elif int(err_["value"]) > 400:
    #     logger.error("There is something wrong in nbox. Raise an issue on github: https://github.com/NimbleBoxAI/nbox/issues")
    # logger.error(err_["grpc_message"])
    logger.error(e)
  else:
    return resp
  if raise_on_error:
    logger.error("NBX-RPC error, see above for details")
    raise


def streaming_rpc(stub: callable, message, err_msg: str, raise_on_error: bool = True):
  """convienience function for streaming from a gRPC stub"""
  try:
    data_iter = stub(message)
    for data in data_iter:
      yield data
  except RpcError as e:
    logger.error(err_msg)
    logger.error(e.details())
    if raise_on_error:
      raise e

def message_to_json(message):
  """convert message to json"""
  return MessageToJson(
    message=message,
    including_default_value_fields=True,
    preserving_proto_field_name=True,
    indent=2,
    sort_keys=False,
    use_integers_for_enums=True,
    float_precision=4
  )

def message_to_dict(message):
  """convert message to dict"""
  return MessageToDict(
    message = message,
    including_default_value_fields=True,
    preserving_proto_field_name=True,
    use_integers_for_enums=True,
    float_precision=4
  )

def dict_to_message(dict, message):
  """load dict into message"""
  ParseDict(dict, message)
  return message


def write_string_to_file(message: message.Message, file):
  from google.protobuf.text_format import PrintMessage
  with open(file, "w") as f:
    PrintMessage(message, f)

def read_file_to_string(file, message: message.Message):
  from google.protobuf.text_format import Parse
  with open(file, "r") as f:
    Parse(f.read(), message)
  return message

def write_binary_to_file(message: message.Message, file):
  with open(file, "wb") as f:
    f.write(message.SerializeToString())

def read_file_to_binary(file, message: message.Message):
  with open(file, "rb") as f:
    message.ParseFromString(f.read())
  return message


################################################################################
"""
# gRPC helpers

The functions defined below are from 
"""
################################################################################

def from_any_pb(pb_type, any_pb):
  """Converts an ``Any`` protobuf to the specified message type.
  Args:
    pb_type (type): the type of the message that any_pb stores an instance
      of.
    any_pb (google.protobuf.any_pb2.Any): the object to be converted.
  Returns:
    pb_type: An instance of the pb_type message.
  Raises:
    TypeError: if the message could not be converted.
  """
  msg = pb_type()

  # Unwrap proto-plus wrapped messages.
  if callable(getattr(pb_type, "pb", None)):
    msg_pb = pb_type.pb(msg)
  else:
    msg_pb = msg

  # Unpack the Any object and populate the protobuf message instance.
  if not any_pb.Unpack(msg_pb):
    raise TypeError(
      "Could not convert {} to {}".format(
        any_pb.__class__.__name__, pb_type.__name__
      )
    )

  # Done; return the message.
  return msg

def check_oneof(**kwargs):
  """Raise ValueError if more than one keyword argument is not ``None``.
  Args:
    kwargs (dict): The keyword arguments sent to the function.
  Raises:
    ValueError: If more than one entry in ``kwargs`` is not ``None``.
  """
  # Sanity check: If no keyword arguments were sent, this is fine.
  if not kwargs:
    return

  not_nones = [val for val in kwargs.values() if val is not None]
  if len(not_nones) > 1:
    raise ValueError(
      "Only one of {fields} should be set.".format(
        fields=", ".join(sorted(kwargs.keys()))
      )
    )

def get_messages(module):
  """Discovers all protobuf Message classes in a given import module.
  Args:
    module (module): A Python module; :func:`dir` will be run against this
      module to find Message subclasses.
  Returns:
    dict[str, google.protobuf.message.Message]: A dictionary with the
      Message class names as keys, and the Message subclasses themselves
      as values.
  """
  answer = collections.OrderedDict()
  for name in dir(module):
    candidate = getattr(module, name)
    if inspect.isclass(candidate) and issubclass(candidate, message.Message):
      answer[name] = candidate
  return answer

def _resolve_subkeys(key, separator="."):
  """Resolve a potentially nested key.
  If the key contains the ``separator`` (e.g. ``.``) then the key will be
  split on the first instance of the subkey::
    >>> _resolve_subkeys('a.b.c')
    ('a', 'b.c')
    >>> _resolve_subkeys('d|e|f', separator='|')
    ('d', 'e|f')
  If not, the subkey will be :data:`None`::
    >>> _resolve_subkeys('foo')
    ('foo', None)
  Args:
    key (str): A string that may or may not contain the separator.
    separator (str): The namespace separator. Defaults to `.`.
  Returns:
    Tuple[str, str]: The key and subkey(s).
  """
  parts = key.split(separator, 1)

  if len(parts) > 1:
    return parts
  else:
    return parts[0], None

def get(msg_or_dict, key, default=_SENTINEL):
  """Retrieve a key's value from a protobuf Message or dictionary.
  Args:
    mdg_or_dict (Union[~google.protobuf.message.Message, Mapping]): the
      object.
    key (str): The key to retrieve from the object.
    default (Any): If the key is not present on the object, and a default
      is set, returns that default instead. A type-appropriate falsy
      default is generally recommended, as protobuf messages almost
      always have default values for unset values and it is not always
      possible to tell the difference between a falsy value and an
      unset one. If no default is set then :class:`KeyError` will be
      raised if the key is not present in the object.
  Returns:
    Any: The return value from the underlying Message or dict.
  Raises:
    KeyError: If the key is not found. Note that, for unset values,
      messages and dictionaries may not have consistent behavior.
    TypeError: If ``msg_or_dict`` is not a Message or Mapping.
  """
  # We may need to get a nested key. Resolve this.
  key, subkey = _resolve_subkeys(key)

  # Attempt to get the value from the two types of objects we know about.
  # If we get something else, complain.
  if isinstance(msg_or_dict, message.Message):
    answer = getattr(msg_or_dict, key, default)
  elif isinstance(msg_or_dict, collections.abc.Mapping):
    answer = msg_or_dict.get(key, default)
  else:
    raise TypeError(
      "get() expected a dict or protobuf message, got {!r}.".format(
        type(msg_or_dict)
      )
    )

  # If the object we got back is our sentinel, raise KeyError; this is
  # a "not found" case.
  if answer is _SENTINEL:
    raise KeyError(key)

  # If a subkey exists, call this method recursively against the answer.
  if subkey is not None and answer is not default:
    return get(answer, subkey, default=default)

  return answer

def _set_field_on_message(msg, key, value):
  """Set helper for protobuf Messages."""
  # Attempt to set the value on the types of objects we know how to deal
  # with.
  if isinstance(value, (collections.abc.MutableSequence, tuple)):
    # Clear the existing repeated protobuf message of any elements
    # currently inside it.
    while getattr(msg, key):
      getattr(msg, key).pop()

    # Write our new elements to the repeated field.
    for item in value:
      if isinstance(item, collections.abc.Mapping):
        getattr(msg, key).add(**item)
      else:
        # protobuf's RepeatedCompositeContainer doesn't support
        # append.
        getattr(msg, key).extend([item])
  elif isinstance(value, collections.abc.Mapping):
    # Assign the dictionary values to the protobuf message.
    for item_key, item_value in value.items():
      set(getattr(msg, key), item_key, item_value)
  elif isinstance(value, message.Message):
    getattr(msg, key).CopyFrom(value)
  else:
    setattr(msg, key, value)

def set(msg_or_dict, key, value):
  """Set a key's value on a protobuf Message or dictionary.
  Args:
    msg_or_dict (Union[~google.protobuf.message.Message, Mapping]): the
      object.
    key (str): The key to set.
    value (Any): The value to set.
  Raises:
    TypeError: If ``msg_or_dict`` is not a Message or dictionary.
  """
  # Sanity check: Is our target object valid?
  if not isinstance(msg_or_dict, (collections.abc.MutableMapping, message.Message)):
    raise TypeError(
      "set() expected a dict or protobuf message, got {!r}.".format(
        type(msg_or_dict)
      )
    )

  # We may be setting a nested key. Resolve this.
  basekey, subkey = _resolve_subkeys(key)

  # If a subkey exists, then get that object and call this method
  # recursively against it using the subkey.
  if subkey is not None:
    if isinstance(msg_or_dict, collections.abc.MutableMapping):
      msg_or_dict.setdefault(basekey, {})
    set(get(msg_or_dict, basekey), subkey, value)
    return

  if isinstance(msg_or_dict, collections.abc.MutableMapping):
    msg_or_dict[key] = value
  else:
    _set_field_on_message(msg_or_dict, key, value)

def setdefault(msg_or_dict, key, value):
  """Set the key on a protobuf Message or dictionary to a given value if the
  current value is falsy.
  Because protobuf Messages do not distinguish between unset values and
  falsy ones particularly well (by design), this method treats any falsy
  value (e.g. 0, empty list) as a target to be overwritten, on both Messages
  and dictionaries.
  Args:
    msg_or_dict (Union[~google.protobuf.message.Message, Mapping]): the
      object.
    key (str): The key on the object in question.
    value (Any): The value to set.
  Raises:
    TypeError: If ``msg_or_dict`` is not a Message or dictionary.
  """
  if not get(msg_or_dict, key, default=None):
    set(msg_or_dict, key, value)

def field_mask(original, modified) -> field_mask_pb2.FieldMask:
  """Create a field mask by comparing two messages.
  Args:
    original (~google.protobuf.message.Message): the original message.
      If set to None, this field will be interpretted as an empty
      message.
    modified (~google.protobuf.message.Message): the modified message.
      If set to None, this field will be interpretted as an empty
      message.
  Returns:
    google.protobuf.field_mask_pb2.FieldMask: field mask that contains
    the list of field names that have different values between the two
    messages. If the messages are equivalent, then the field mask is empty.
  Raises:
    ValueError: If the ``original`` or ``modified`` are not the same type.
  """
  if original is None and modified is None:
    return field_mask_pb2.FieldMask()

  if original is None and modified is not None:
    original = copy.deepcopy(modified)
    original.Clear()

  if modified is None and original is not None:
    modified = copy.deepcopy(original)
    modified.Clear()

  if type(original) != type(modified):
    raise ValueError(
      "expected that both original and modified should be of the "
      'same type, received "{!r}" and "{!r}".'.format(
        type(original), type(modified)
      )
    )

  return field_mask_pb2.FieldMask(paths=_field_mask_helper(original, modified))

def _field_mask_helper(original, modified, current=""):
  answer = []

  for name in original.DESCRIPTOR.fields_by_name:
    field_path = _get_path(current, name)

    original_val = getattr(original, name)
    modified_val = getattr(modified, name)

    if _is_message(original_val) or _is_message(modified_val):
      if original_val != modified_val:
        # Wrapper types do not need to include the .value part of the
        # path.
        if _is_wrapper(original_val) or _is_wrapper(modified_val):
          answer.append(field_path)
        elif not modified_val.ListFields():
          answer.append(field_path)
        else:
          answer.extend(
            _field_mask_helper(original_val, modified_val, field_path)
          )
    else:
      if original_val != modified_val:
        answer.append(field_path)

  return answer

def _get_path(current, name):
  # gapic-generator-python appends underscores to field names
  # that collide with python keywords.
  # `_` is stripped away as it is not possible to
  # natively define a field with a trailing underscore in protobuf.
  # APIs will reject field masks if fields have trailing underscores.
  # See https://github.com/googleapis/python-api-core/issues/227
  name = name.rstrip("_")
  if not current:
    return name
  return "%s.%s" % (current, name)

def _is_message(value):
  return isinstance(value, message.Message)

def _is_wrapper(value):
  return type(value) in _WRAPPER_TYPES



"""
Shell
=====

This module contains the Operators that can be used to interact with the shell.
"""

from nbox import Operator

class ShellCommand(Operator):
  def __init__(self, *commands):
    """Run multiple shell commands, uses ``shelex`` to prevent injection"""
    super().__init__()
    import string

    self.commands = commands
    all_in = []
    for c in self.commands:
      all_in.extend([tup[1] for tup in string.Formatter().parse(c) if tup[1] is not None])

  def forward(self, *args, **kwargs):
    import shlex
    import subprocess

    for comm in self.commands:
      comm = comm.format(*args, **kwargs)
      comm = shlex.split(comm)
      called_process: subprocess.CompletedProcess = subprocess.run(comm, check = True)
      if called_process.returncode != 0:
        raise Exception(f"Command {comm} failed with return code {called_process.stdout}")


class Python(Operator):
  def __init__(self, func, *args, **kwargs):
    """Convert a python function into an operator, everything has to be passed at runtime"""
    super().__init__()
    self.fak = (func, args, kwargs)

  def forward(self):
    return self.fak[0](*self.fak[1], **self.fak[2])


class PythonScript(Operator):
  def __init__(self, fpath, python_path = "/usr/bin/python3", **kwargs):
    """Run any file in python as an operator"""
    super().__init__()
    self.fp = fpath
    self.python_path = python_path
    self.kwargs = kwargs

    self.kwargs_strings = []
    for k,v in kwargs.items():
      if v != None:
        self.kwargs_strings.append(f"--{k}={v}")
      else:
        self.kwargs_strings.append(f"--{k}")
    kwargs = "\ \n".join(self.kwargs_strings)
    self.com = ShellCommand(f"{python_path} {fpath} {kwargs}")

  def forward(self):
    return self.com()

class NBToScript(Operator):
  def __init__(self, notebook_path: str, py_file: str) -> None:
    super().__init__()
    self.nb = notebook_path
    self.trg = py_file

  def forward(self):
    import json

    with open(self.nb, "r") as f:
      data = json.load(f)
    
    code_script = "\n# This is an Autogenerated file by NBToScript Operators\n# Find out more: nimblebox.ai\n"
    # here we can start iterating over the code cells and fill up a script
    for cell in data["cells"]:
      if cell["cell_type"] == "code":
        code_script += "".join(cell["source"]) + "\n"

    with open(self.trg, "w") as f:
      f.write(code_script)

class GitClone(Operator):
  def __init__(self, url, path = None, branch = None):
    """Clone a git repository into a path"""
    super().__init__()
    self.url = url
    self.path = path
    self.branch = branch

  def forward(self):
    # git clone -b <branch> <repo> <path>
    import subprocess
    command = ["git", "clone"]
    if self.branch:
      command.append("-b")
      command.append(self.branch)
    if self.path:
      command.append(self.path)
    command.append(self.url)
    subprocess.run(command, check = True)




# weird file name, yessir. Why? because
# from nbox.framework.on_functions import PureFunctionParser
# read as "from nbox's framework on Functions import the pure-function Parser"

"""
This contains the code for our public static code parsers. We will be moving them internally as we expand our
scope. Till then here's all the code we have written till now. Keep a lookout for ``Timeless`` from NimbleBox.
"""

import ast
import base64
import inspect
from typing import Union
from uuid import uuid4

from ..utils import logger
from ..hyperloop.dag_pb2 import DAG, Flowchart, Node, Code, Edge, RunStatus

# ==================


# ==================
# the next set of functions are meant as support methods to create nbox_strings IR.

class NboxStrings:
  """Class to generate all the strings, for our representation."""
  OP_TO_STRING = {
    "function": "FUNCTION: {name} ( {inputs} ) => [ {outputs} ]",
    "define": "DEFINE: {name} ( {inputs} )",
    "for": "FOR: ( {iter} ) => [ {target} ]",
    "return": "RETURN: [ {value} ]",
  }

  def __init__(self):
    pass

  def function(self, name, inputs, outputs):
    """Render code for any function."""
    return self.OP_TO_STRING["function"].format(
      name=name,
      inputs=", ".join([f"{x['kwarg']}={x['value']}" for x in inputs]),
      outputs=", ".join(outputs)
    )

  def define(self, name, inputs):
    """Render code for any definition."""
    return self.OP_TO_STRING["define"].format(
      name=name,
      inputs=", ".join([f"{x['kwarg']}={x['value']}" for x in inputs])
    )

  def for_loop(self, iter, target):
    """Render code for any for loop."""
    return self.OP_TO_STRING["for"].format(
      iter=iter,
      target=", ".join(target)
    )

  def return_statement(self, value):
    """Render code for any return statement."""
    return self.OP_TO_STRING["return"].format(
      value=value
    )

nbxl = NboxStrings()

def write_program(nodes):
  """Render code for any program."""
  for i, n in enumerate(nodes):
    logger.debug(f"{i:03d}|{n.get('nbox_string', n.get('info').get('nbox_string'))}")


# ==================

def get_code_portion(cl, lineno, col_offset, end_lineno, end_col_offset, b64 = True, **_):
  """Returns code piece for any section given all the lines it contains"""
  sl, so, el, eo = lineno, col_offset, end_lineno, end_col_offset
  if sl == el:
    return cl[sl-1][so:eo]
  code = ""
  for i in range(sl - 1, el, 1):
    if i == sl - 1:
      code += cl[i][so:]
    elif i == el - 1:
      code += "\n" + cl[i][:eo]
    else:
      code += "\n" + cl[i]
  
  # convert to base64
  if b64:
    return base64.b64encode(code.encode()).decode()
  return code

def parse_args(node: ast.arguments):
  """Takes an ``ast.arguments`` list and reutrns kwarg/value pair from it."""
  inputs = []
  for a in node.args:
    a = a.arg if isinstance(a, ast.arg) else a
    inputs.append({
      "kwarg": None,
      "value": a,
    })
  for a in node.kwonlyargs:
    inputs.append({
      "kwarg": a[0],
      "value": a[1],
    })
  if node.vararg:
    inputs.append({
      "kwarg": "*"+node.vararg.arg,
      "value": None
    })
  if node.kwarg:
    inputs.append({
      "kwarg": "**"+node.kwarg.arg,
      "value": None
    })
  return inputs

def get_name(node):
  """Given a node, try to extract it's name."""
  if isinstance(node, ast.Name):
    return node.id
  elif isinstance(node, ast.Constant):
    return node.value
  elif isinstance(node, ast.Attribute):
    return get_name(node.value) + "." + node.attr
  elif isinstance(node, ast.Call):
    return get_name(node.func)

def parse_kwargs(node, lines):
  """Given a node, try to extract it's kwargs."""
  if isinstance(node, ast.Name):
    return node.id
  elif isinstance(node, ast.Constant):
    val = node.value
    return val
  elif isinstance(node, ast.keyword):
    arg = node.arg
    value = node.value
    if 'id' in value.__dict__:
      # arg = my_model
      return (arg, value.id)
    elif 'value' in value.__dict__:
      if isinstance(value.value, ast.Call):
        # arg = my_model(...)
        return (arg, get_code_portion(lines, b64 = False, **value.value.__dict__))
      else:
        # arg = 20
        return (arg, value.value)
    elif 'func' in value.__dict__:
      #   arg = some_function(with, some=args)
      #   ^^^   ^^^^^
      # kwarg   value
      return {"kwarg": arg, "value": get_code_portion(lines, b64 = False, **value.__dict__)}
  elif isinstance(node, ast.Call):
    return get_code_portion(lines, **node.func.__dict__)

def node_assign_or_expr(node, lines, node_proto: Node) -> Union[Node, None]:
  """Return ``nbox.hyperloop`` Node for assign or expressions"""
  # print(get_code_portion(lines, b64 = False, **node.__dict__))
  value = node.value
  try:
    name = get_name(value.func)
  except AttributeError:
    return None
  args = [parse_kwargs(x, lines) for x in value.args + value.keywords]
  inputs = []
  for a in args:
    if isinstance(a, dict):
      inputs.append(a)
      continue
    inputs.append({
      "kwarg": a[0] if isinstance(a, tuple) else None,
      "value": a[1] if isinstance(a, tuple) else a,
    })

  outputs = []
  if isinstance(node, ast.Assign):
    targets = node.targets[0]
    outputs = [parse_kwargs(x, lines) for x in targets.elts] \
      if isinstance(targets, ast.Tuple) \
      else [parse_kwargs(targets, lines)
    ]

  nbox_string = nbxl.function(name, inputs, outputs)

  node_proto.MergeFrom(Node(
    name = name,
    type = Node.NodeType.OP,
    info = Code(
      name = name,
      nbox_string = nbox_string,
      code = get_code_portion(lines, **node.__dict__),
      lineno = node.lineno,
      col_offset = node.col_offset,
      end_lineno = node.end_lineno,
      end_col_offset = node.end_col_offset,
      inputs = {str(x["kwarg"]): str(x["value"]) for x in inputs},
      outputs = {x: "" for x in outputs},
    )
  ))
  return node_proto

def node_if_expr(node: ast.IfExp, lines: list, node_proto: Node) -> Node:
  """Return ``nbox.hyperloop`` Node for if conditional"""
  def get_conditions(node, lines, conds = []):
    if not hasattr(node, "test"):
      else_cond = list(filter(lambda x: x["condition"] == "else", conds))
      if not else_cond:
        conds.append({
          "condition": "else",
          "code": dict(
            lineno = node.lineno,
            col_offset = node.col_offset,
            end_lineno = node.end_lineno,
            end_col_offset = node.end_col_offset,
          )
        })
      else:
        cond = else_cond[0]
        cond["code"]["end_lineno"] = node.end_lineno
        cond["code"]["end_col_offset"] = node.end_col_offset
    else:
      condition = get_code_portion(lines, **node.test.__dict__)

      # need to run this last or "else" comes up first
      conds.append({
        "condition": condition,
        "code": {
          "lineno": node.lineno,
          "col_offset": node.col_offset,
          "end_lineno": node.end_lineno,
          "end_col_offset": node.end_col_offset,
        }
      })
      for x in node.orelse:
        get_conditions(x, lines, conds)

    return conds

  all_conditions = get_conditions(node, lines, conds = [])
  ends = []
  for b0, b1  in zip(all_conditions[:-1], all_conditions[1:]):
    ends.append([b0["code"], b1["code"]])
  for i in range(len(ends)):
    ends[i] = {
      "lineno": ends[i][0]["lineno"],
      "col_offset": ends[i][0]["col_offset"],
      "end_lineno": ends[i][1]["lineno"],
      "end_col_offset": ends[i][1]["col_offset"],
    }
  ends += [all_conditions[-1]["code"]]

  conditions = []
  for i, c in enumerate(all_conditions):
    box = ends[i]
    _node = Code(
      name = f"if-{i}",
      nbox_string = c["condition"],
      lineno = box['lineno'],
      col_offset = box['col_offset'],
      end_lineno = box['end_lineno'],
      end_col_offset = box['end_col_offset'],
      inputs = [],
      outputs = [],
    )
    conditions.append(_node)

  # update the node_proto
  node_proto.MergeFrom(Node(
    name = "if",
    type = Node.NodeType.BRANCHING,
    info = Code(
      name = "if",
      nbox_string = "IF: { " + ", ".join(x.nbox_string for x in conditions) + " }",
      lineno = node.lineno,
      col_offset = node.col_offset,
      end_lineno = node.end_lineno,
      end_col_offset = node.end_col_offset,
      inputs = {},
      outputs = {},
      conditions = conditions,
    )
  ))
  return node_proto

def node_for_expr(node: ast.For, lines: list, node_proto: Node) -> Node:
  """Return ``nbox.hyperloop`` Node for as for loop."""
  iter_str = get_code_portion(lines, **node.iter.__dict__)
  if isinstance(node.target, ast.Tuple):
    targets = [x.id for x in node.target.elts]
  else:
    targets = [get_name(node.target)]
  code_ = get_code_portion(lines, **node.__dict__)
  nbox_string = nbxl.for_loop(iter_str, targets)

  # update node_proto
  node_proto.MergeFrom(Node(
    name = "for",
    type = Node.NodeType.LOOP,
    info = Code(
      name = "for",
      code = code_,
      nbox_string = nbox_string,
      lineno = node.lineno,
      col_offset = node.col_offset,
      end_lineno = node.end_lineno,
      end_col_offset = node.end_col_offset,
      inputs = {},
      outputs = {'None':x for x in targets},
    )
  ))
  return node_proto

def node_return(node: ast.Return, lines, node_proto: Node) -> Node:
  """Return ``nbox.hyperloop`` Node for a return method."""
  if isinstance(node.value, ast.Tuple):
    returns = [x.id for x in node.value.elts]
  else:
    returns = [get_name(node.value)]
  node_proto.MergeFrom(Node(
    name = "return",
    # type = Node.NodeType.RETURN,
    info = Code(
      name = "return",
      code = get_code_portion(lines, **node.__dict__),
      nbox_string = nbxl.return_statement(returns),
      lineno = node.lineno,
      col_offset = node.col_offset,
      end_lineno = node.end_lineno,
      end_col_offset = node.end_col_offset,
      inputs = {},
      outputs = {'None':'None'},
    )
  ))
  return node_proto

def def_func_or_class(node, lines, node_proto: Node) -> Node:
  """Return ``nbox.hyperloop`` Node for any function or class definition."""
  node_proto.MergeFrom(Node(
    name = f"symbol-{node.name}",
    type = Node.NodeType.OP, # TODO: @yashbonde expand node types to include class/func definitions
    info = Code(
      name = "return",
      code = get_code_portion(lines, **node.__dict__),
      nbox_string = "",
      lineno = node.lineno,
      col_offset = node.col_offset,
      end_lineno = node.end_lineno,
      end_col_offset = node.end_col_offset,
      inputs = {},
      outputs = {'None':'None'},
    )
  ))
  return node_proto

def async_def_func_or_class(node, lines, node_proto: Node) -> Node:
  """Return ``nbox.hyperloop`` Node for any asyn function."""
  print(node.__dict__)
  node_proto.MergeFrom(Node(
    name = f"symbol-{node.name}",
    type = Node.NodeType.OP, # TODO: @yashbonde expand node types to include class/func definitions
    info = Code(
      name = "return",
      code = get_code_portion(lines, **node.__dict__),
      nbox_string = "",
      lineno = node.lineno,
      col_offset = node.col_offset,
      end_lineno = node.end_lineno,
      end_col_offset = node.end_col_offset,
      inputs = {},
      outputs = {'None':'None'},
    )
  ))
  return node_proto


# ==================

type_wise_logic = {
  # defns ------
  ast.FunctionDef: def_func_or_class,
  ast.ClassDef: def_func_or_class,
  ast.AsyncFunctionDef: def_func_or_class,

  # nodes ------
  ast.Assign: node_assign_or_expr,
  ast.Expr: node_assign_or_expr,
  ast.If: node_if_expr,
  ast.For: node_for_expr,

  # Return ------
  ast.Return: node_return,

  # todos ------
  # ast.AsyncFunctionDef: async_func_def,
  # ast.Await: node_assign_or_expr,
}

# ==================


def code_node(execution_index, expr, code_lines) -> Node:
  """Return default ``nbox.hyperloop`` Node."""
  # code pieces that are not yet supported should still see the code
  return Node(
    id = str(uuid4()),
    execution_index = execution_index,
    name = f"codeblock-{execution_index}",
    type = Node.NodeType.OP,
    operator = "CodeBlock",
    nbox_string = f"CODE: {str(type(expr))}", # :str
    run_status = RunStatus(), # no need to initialise this object
    info = Code(
      name = f"codeblock-{execution_index}", # :str
      code = get_code_portion(code_lines, bs64 = True, **expr.__dict__), # :str (base64)
      nbox_string = None, # :str
      lineno = expr.lineno, # :int
      col_offset = expr.col_offset, # :int
      end_lineno = expr.end_lineno, # :int
      end_col_offset = expr.end_col_offset, # :int
      inputs = [], # :list[Dict[Any, Any]]
      outputs = [], # :list[str]
    )
  )


def get_nbx_flow(forward) -> DAG:
  """Get NBX flowchart. Read python grammar `here <https://docs.python.org/3/reference/grammar.html>`_

  Args:
      forward (callable): the function whose flowchart is to be generated
  """
  if type(forward) == str:
    code = forward
  else:
    code = inspect.getsource(forward).strip() # get code string from operator
  code_lines = code.splitlines()
  ast_node = ast.parse(code)

  edges = {} # this is the flow
  nodes = {} # this is the operators
  symbols_to_nodes = {} # this is things that are defined at runtime

  try:
    iter_obj = ast_node.body[0].body if hasattr(ast_node.body[0], 'body') else ast_node.body
    for i, expr in enumerate(iter_obj):
      # create the empty node that will be used everywhere
      if not type(expr) in type_wise_logic:
        node: Node = code_node(i, expr, code_lines)
        nodes[node.id] = node
        continue

      # define an initial proto and then then other functions will fill it up
      node_proto = Node(id = str(uuid4()), execution_index = i)
      output = type_wise_logic[type(expr)](expr, code_lines, node_proto)
      if output is None:
        continue
      nodes[node_proto.id] = node_proto

    # edges for execution order can be added
    _node_ids = tuple(nodes.keys())
    for op0, op1 in zip(_node_ids[:-1], _node_ids[1:]):
      _id = f"edge-{op0}-X-{op1}"
      edges[_id] = Edge(
        id = _id,
        source = op0,
        target = op1,
        type = Edge.EdgeType.EXECUTION_ORDER,
        nbox_string = None
      )
  except Exception as e:
    logger.error("Error in parsing the job-flow")
    logger.error(e)
    nodes = {}
    edges = {}
    symbols_to_nodes = {}

  return DAG(flowchart = Flowchart(nodes = nodes, edges = edges), symbols=symbols_to_nodes)
